/*
 *	Copyright (C) 2007, 2008, 2009. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVIP_H
#define QVIP_H

#include <iostream>

#include <qvdefines.h>
#include <QVImage>
#include <QVPolyline>

/// @ingroup qvip
/// @todo document this
void FilterHarrisCornerResponseImage(const QVImage<uChar> &image, QVImage<sFloat> &result, int aperture=3, int avgwindow=5, const QPoint &destROIOffset = QPoint(0,0));

/// @ingroup qvip
/// @todo document this
void FilterDoG(const QVImage<uChar> &image, QVImage<uChar> &result);

/// @ingroup qvip
/// @deprecated see @ref FilterHessianCornerResponseImage
/// @todo document this
void SobelCornerResponseImage(const QVImage<sFloat> &image, QVImage<sFloat> &result);

/// @ingroup qvip
/// @todo document this
void FilterHessianCornerResponseImage(	const QVImage<sFloat> &image, QVImage<sFloat> &result,
					const QPoint &destROIOffset = QPoint(0,0));

/// @ingroup qvip
/// @todo document this
void FilterNormalize(const QVImage<uChar,1> &image, QVImage<uChar,1> &equalized, const QPoint &destROIOffset = QPoint(0,0));

/// @ingroup qvip
/// @todo document this
void FilterNormalize(const QVImage<sFloat,1> &image, QVImage<sFloat,1> &equalized, const QPoint &destROIOffset = QPoint(0,0));

/// @ingroup qvip
class QVComponentTree;
void FilterPruneComponentTreeSmallRegions(QVImage<uChar> &image, QVComponentTree &componentTree, uInt area);

#include <qvmath/qvvector.h>
#include <QPoint>
/// @ingroup qvip
void FilterSeparable(const QVImage<sFloat, 1> &image, QVImage<sFloat, 1> &dest,
	const QVVector &rowFilter, const QVVector &colFilter, const QPoint &destROIOffset = QPoint(0,0));
/*!
@brief Generates a binary image of local grey-scale maximal value pixels.
@ingroup qvip

This function receives a QVImage, and generates a binary image where each pixel is set to IPP_MAX_8U if the pixel in the
original image is strict maximal in value regarding to pixels in a vicinity window of colMaskSize width, and rowMaskSize
height.

@todo
	- Fix resulting image ROI, obtain maximums restricted to the ROI.
	- Use IPP functions to obtain the result.

@param src source image.
@param dest binary image that will contain maximal values.
@param colMaskSize width of the vicinity.
@param rowMaskSize height of the vicinity.
*/
void FilterLocalMax(const QVImage<sFloat> &src, QVImage<uChar> &dest, uInt colMaskSize, uInt rowMaskSize, sFloat threshold = 0);

/// @todo document this
/// @ingroup qvip
QMap<sFloat, QPointF> maximalPoints(const QVImage<sFloat> &image, const double threshold = 1.0, const int windowRadius = 2);

/// @todo document this
/// @ingroup qvip
QMap<sFloat, QPointF>  fastMaximalPoints(const QVImage<sFloat> &image, const double threshold = 1.0, const int windowSize = 2);

/// @brief Sorts pixels in an image, given their gray-scale value.
/// @ingroup qvip
///
/// This function obtains a QVector, that maps any gray-scale values (0..255) to the set of QPoints that hold that gray-scale
/// value, in a given image. It can be used for very fast pixel sorting.
///
/// The function itself is based in \htmlonly <a href="http://en.wikipedia.org/wiki/Counting_sort">Counting Sort algorithm</a> \endhtmlonly, so the cost of this function is O(number_of_columns * number_of_rows).
///
/// @param img source image to obtain sorted lists of QPoints.
/// @return A QVector, indexed by the gray-scale value (0-255), to a QVector of QPoints, containing the corresponding points to the pixels that hold that gray-scale value in the image.
QVector< QVector< QPoint > > CountingSort(const QVImage<uChar, 1> &img);

/*!
@brief Simplifies a contour eliminating points of little area using IPE algorithm
@ingroup qvip

This function eliminates points of a QVPolylineF, simplificating it.
Points are recursively eliminated while a) their distance to the line joining
its two immediate neighbours is smaller than a given maximum value, or
b) the number of points of the approximated polyline falls below a given
number (depending on the chosen method).

The return value (which indicates the cost of the first not deleted point)
and the optional parameter max_removed_cost are useful to estimate
a measure of how well the polyline approximation did. A good
approximation will have (relative) low max_removed_cost/return_cost ratio,
and (absolute) low max_removed_cost value.

@param polyline polyline to simplify.
@param result will store resulting simplified polyline.
@param param maximal distance for a point to eliminate it, if
             maxNumberOfPointsMethod is FALSE, or maximal number of points in the
             result polyline, if maxNumberOfPointsMethod is TRUE.
@param maxNumberOfPointsMethod if TRUE, the procedure ends when the number
             of points of the resulting polyline is lesser or equal to param.
             If FALSE, the procedure ends when there are no more points
             with distance to line given by its neighbours smaller than
             param.
@param intersectLines If TRUE, a post-processing stage adjusts final
             points using all the deleted points among vertexs to fit
             straight lines, whose intersection by pairs gives the final
             points (recommended for greater precision, at a minimal
             additional cost).
@param max_removed_cost If not NULL, pointer to a float value that the
             procedure will fill with the maximum cost (distance to
             corresponding line) of all the removed points.
@return The function returns the cost value of the first NOT deleted point
        of the polylyne.
@see IterativePointElimination(const QVPolylineF &, QVPolyline &, const double, bool, bool, double *);
*/
double IterativePointElimination(const QVPolyline &polyline, QVPolyline &result,
	const double param, bool maxNumberOfPointsMethod=FALSE,
	bool intersectLines=TRUE, double *max_removed_cost=NULL);

/*!
@brief Simplifies a contour eliminating points of little area using IPE algorithm
@ingroup qvip

This is an overloaded version of the <i>IterativePointElimination</i>, provided for convenience.

@see IterativePointElimination(const QVPolyline &, QVPolyline &, const double, bool, bool, double *);
*/
double IterativePointElimination(const QVPolylineF &polyline, QVPolylineF &result,
	const double param, bool maxNumberOfPointsMethod=FALSE,
	bool intersectLines=TRUE, double *max_removed_cost=NULL);

/*!
@brief Obtains the border contour of a connected set of pixels in an image, given a membership condition.
@ingroup qvip

A contour is a sequence of pixels from an image, were every two contiguous pixels in the sequence are neighbours in
the image.

This function obtains the contour that depicts the border set of a <a href="http://en.wikipedia.org/wiki/Connected_set">connected set</a>
of pixels in an image, given a membership condition.

For this border contour extractor function, the membership condition will be that pixels will be considered to be members of
a connected if and only if their gray-scale value is equal or greater than a given threshold value.

The function implements an automaton, that will receive an image and a starting point in its arguments, and will look for 
the closest border pixel to that point.

That border pixel will have a gray-level value equal or higher than a threshold value, given also in the arguments of the
function, and will transverse the border of the set, appending every point that it finds in it in a @ref QVPolyline object,
that will be the returning value of the function.

Border contours obtained with this function can be either inner borders (corresponding to the border of an empty space inside
the connected set) or the outer border of the connected set. The programer can tell wether a given border is inner or outer
by the boolean value @ref QVPolyline::direction, which will store <i>TRUE</i> for an outer contour, and <i>FALSE</i> for an
inner contour.

@param image image from were to obtain the poly-line.
@param point any point inside the connected set.
@param threshold threshold to consider pixels inside or outside the connected set.
@returns a QVPolyline that follows the border set of the connected set.
@see getConnectedSetBorderContoursThreshold
*/
QVPolyline getConnectedSetBorderContourThreshold(const QVImage<uChar> &image, const QPoint point, const uChar threshold = 128);

/*!
@brief Obtains a list of the border contours of the connected sets in an image, according to a membership condition.
@ingroup qvip

This function gets the border contours for all of the <a href="http://en.wikipedia.org/wiki/Connected_set">connected sets</a>
of pixels in an image, using the function @ref getConnectedSetBorderContourThreshold(const QVImage<uChar> &, const QPoint, const uChar), and returns them in a list of polylines (QList<QVPolyline>).

Thus, it considers every point inside a connected set, if it has a gray-level equal or higher than a given value. An example of
the contours obtained is depicted in the next image:

\image html bordercontours.png

Circles indicate the first of the pixels in each border set, and following pixels are joined with a continuous line, until
the last of the pixels in the set, depicted with an asterisk. 

@param image image from were to obtain the poly-lines.
@param threshold threshold to consider pixels inside or outside the connected sets.
@returns a QList of QVPolyline's that follows the border sets of the connected sets in the image.
@see getConnectedSetBorderContourThreshold
*/
QList<QVPolyline> getConnectedSetBorderContoursThreshold(const QVImage <uChar> &image, const uChar threshold = 128);

/*!
@brief Obtains a list of the 4-connected contour lines in the image
@ingroup qvip

A contour is a sequence of pixels from an image, were every two contiguous pixels in the sequence are neighbours in
the image.

This function can be used to obtain the 4-connected sets depicted in a border response image, obtained with Canny operator,
or other image border response algorithm. It will return the contours as a @ref QVPolyline. Pixels with gray-level
higher than threshold, will be considered as border pixels. An example of the contours obtained is depicted in the next image:

\image html contours4connected.png

Circles indicate the first of the pixels in each set, and following pixels are joined with a continuous line, until the last
of the pixels in the set, depicted with an asterisk. Ambiguous situations (such as pixels with more than two 4-connected
neihgbour pixels) return unpredictable joining results.

@param image image from were to obtain the poly-lines.
@param threshold threshold to test if a pixel is considered a border or not.
@returns the list of contours (as @ref QVPolyline) in the image.
@see Canny
*/
QList<QVPolyline> getLineContoursThreshold4Connectivity(const QVImage<uChar> &image, const uChar threshold = 128);

/*!
@brief Obtains a list of the 8-connected contour lines in the image
@ingroup qvip

A contour is a sequence of pixels from an image, were every two contiguous pixels in the sequence are neighbours in
the image.

This function can be used to obtain the 8-connected sets depicted in a border response image, obtained with Canny operator,
or other image border response algorithm. It will return the contours as a @ref QVPolyline. Pixels with gray-level
higher than threshold, will be considered as border pixels. An example of the contours obtained is depicted in the next image:

\image html contours8connected.png

Circles indicate the first of the pixels in each set, and following pixels are joined with a continuous line, until the last
of the pixels in the set, depicted with an asterisk. Ambiguous situations (such as pixels with more than two 8-connected
neihgbour pixels) return unpredictable joining results.

@param image image from were to obtain the poly-lines.
@param threshold threshold to test if a pixel is considered a border or not.
@returns the list of contours (as @ref QVPolyline) in the image.
@see Canny
*/
QList<QVPolyline> getLineContoursThreshold8Connectivity(const QVImage<uChar> &image, const uChar threshold = 128);

#endif

